import os
import codecs

from django.utils import six
from django.conf import settings
from django.core.management import BaseCommand
from django.core.exceptions import ImproperlyConfigured

from djsw_wrapper.core import Swagger
from djsw_wrapper.utils import Template
from djsw_wrapper.router import SwaggerRouter

class Command(BaseCommand):
    help = "Generates stub controllers for endpoints specified in API scheme (both YAML and JSON are supported)"

    def add_arguments(self, parser):
        parser.add_argument('--name', nargs = '?', default = 'controllers', help = 'Name of module to be created (without extenstion)')
        parser.add_argument('--generate', action = 'store_true', dest = 'generate', help = 'Generate handlers according to spec')

    def handle(self, *args, **options):
        schema = getattr(settings, 'SWAGGER_SCHEMA', None)
        module = getattr(settings, 'SWAGGER_MODULE', None)

        if not schema:
            raise ImproperlyConfigured('You have to provide SWAGGER_SCHEMA setting pointing to desired schema')
        if not module:
            raise ImproperlyConfigured('You have to specify desired controller module name in SWAGGER_MODULE setting')

        router = SwaggerRouter()

        print('Inspecting available controllers...')

        router.update(True)
        router.process()

        print()
        print('Following classes and methods are going to be generated:')

        enum = router.get_enum()

        for name in enum:
            print("{} : {}".format(name, [x['method'] for x in enum[name]['methods']]))

        if(options['generate']):
            template = Template()
            filename = module.split('.')[-1] + '.py'
            structure = [{ 'name' : name, 'data' : data } for name, data in six.iteritems(enum)]

            print('Generating handlers ({})...'.format(filename))

            with codecs.open(filename, 'w', 'utf-8') as f:
                f.write(template.render(template_name = 'view.jinja', names = structure))

            print('Done.')
        else:
            print()
            print('Use --generate option to create them')




import re
import logging
import importlib

from django.utils import six
from django.conf.urls import url as make_url

from djsw_wrapper.utils import Singleton, Template, Resolver
from djsw_wrapper.makers import SwaggerViewMaker, SwaggerRequestMethodMaker, SwaggerViewClass
from djsw_wrapper.params import SwaggerParameter, SwaggerRequestHandler
from djsw_wrapper.errors import SwaggerValidationError, SwaggerGenericError

logger = logging.getLogger(__name__)

class SwaggerRouter(Singleton):
    # name of reference to controller in schema
    cextname = 'x-swagger-router-controller'

    # extract parameters from url path
    paramregex = re.compile(r'\{(\w?[\w\d]*)\}')

    # remove heading and trailing slashes
    wrapregex = re.compile(r'^\/?(.*)\/?')

    # allowed by Swagger 2.0
    allowed_methods = set(['get', 'put', 'post', 'head', 'patch', 'options', 'delete'])

    # return a raw string for url regex
    def makeraw(self, string):
        if six.PY3:
            return string.encode('unicode-escape')
        else:
            return string.encode('string-escape')

    def __init__(self, schema, controllers = None, models = None):
        self.base = schema['basePath']
        self.urls = []
        self.enum = None
        self.paths = schema['paths']
        self.schema = schema
        self.create = False
        self.models = models
        self.handlers = {}
        self.parsers = None
        self.renderers = None
        self.external = None
        self.stubsonly = False
        self.controllers = controllers

        self.process()

    def update(self, create = False):
        self.create = create

    def log(self, *args, **kwargs):
        if self.create:
            print(*args, **kwargs)
        else:
            logger.info(*args, **kwargs)

    # detachable constructor
    def process(self):
        self.stubsonly = False

        # try to import controller module first
        if self.controllers:
            try:
                self.external = importlib.import_module(self.controllers)
            except ImportError:
                self.stubsonly = True

                self.log('Could not import controller module ({}), using stub handlers for all endpoints'.format(str(self.controllers)))
        else:
            self.stubsonly = True

        # enumerate all methods for gen
        self.enum = dict()

        # determine parsers and renderers
        # TODO: do we really need this?
        #
        # APIView.parser_classes = (XMLParser, )
        # self.schema['consumes'] = ...
        # self.schema['produces'] = ...

        # iterate over all paths
        for path in self.paths:
            view = None
            stub = False
            name = None
            controller = None
            child = self.paths[path]

            # get name from schema
            name = child[self.cextname]

            # if we have module, check for controller property and try to use it
            if not self.stubsonly:
                try:
                    controller = getattr(self.external, name)

                except KeyError:
                    stub = True
                    self.log('Controller property for path "{}" is not defined, using stub handler'.format(path))

                except AttributeError:
                    stub = True
                    self.log('Could not find controller "{}" for path "{}", using stub handler'.format(name, path))

            else:
                stub = True

            # pick only allowed methods and make dict of them
            methods = { m : None for m in self.allowed_methods.intersection(set(child)) }

            # check for empty path
            if len(methods) == 0:
                raise SwaggerValidationError('Path "{}" does not contain any supported methods'.format(path))

            # gather all params and wrap them for matching methods
            allparams = set()

            # process methods and responses
            for method in methods:
                responses = child[method].get('responses', None)
                parameters = child[method].get('parameters', None)
                methoddata = { 'params' : None, 'model' : None }

                if parameters:
                    wrapped = list(map(lambda p : SwaggerParameter(p), parameters))
                    methoddata['params'] = wrapped
                    allparams.update([x.name for x in wrapped])

                # TODO: simplify
                if responses:
                    successful = responses[200]
                    schema = successful.get('schema', None)

                    if schema and schema.get('type', None) == 'array':
                        model = Resolver(self.schema, schema['items']['$ref'])
                        mdict = { x : None for x in self.models[model] }
                        methoddata['model'] = [mdict]

                methods[method] = methoddata

            # enumerate named parameters and construct endpoint url
            reg = None
            url = six.moves.urllib.parse.urljoin(self.base, path)
            named = set(self.paramregex.findall(url))

            # if there are any named params, convert 'em to django's format; otherwise just use url
            if len(named):
                reg = re.sub(self.paramregex, r'(?P<\1>[^/.]+)', url) # TODO: make matching length tuneable

                # check if schema missing params described in url
                if not named.issubset(allparams):
                    raise SwaggerValidationError('Path "{}" lacks parameters schema'.format(path))
            else:
                reg = url

            # make regex bounds
            reg = re.sub(self.wrapregex, r'^\1/?$', reg)

            # get documentation (if present)
            doc = child.get('description', None)

            # create fallback name
            tempname = re.sub(self.paramregex, r'', str().join(map(str.capitalize, path.split('/')))) + '_' + str(len(path))

            # create stub view object or use existing controller
            if not self.create:
                if stub:
                    name = tempname
                    view = SwaggerViewMaker(name)
                else:
                    view = controller
            else:
                if not name:
                    name = tempname

                self.enum[name] = { 'methods' : [], 'doc' : doc.splitlines() if doc else None }

            # create serializers
            for method, data in six.iteritems(methods):
                handler = getattr(view, method, None) if not stub else None

                if handler is None:
                    handler = SwaggerRequestMethodMaker(data['model'])

                    if self.create:
                        self.enum[name]['methods'].append({ 'method' : method, 'model' : data['model'] })

                if not self.create:
                    # return validation wrapper if there are some params
                    # or clean (stub) method otherwise
                    wrapped = SwaggerRequestHandler(view, handler, data['params'])

                    # write back to view
                    if stub:
                        view.set_attr(method, wrapped)
                    else:
                        setattr(view, method, wrapped)

            # create doc
            if doc and stub and not self.create:
                view.set_attr('__doc__', doc)

            if not self.create:
                as_view = getattr(view, 'as_view', None)

                # use method views if possible
                if view is not callable and as_view:
                    view = as_view()

                # push to dict
                self.handlers.update({ reg : view })

        # make sorted list and map to django's url()
        if not self.create:
            self.urls = [ make_url(rv[0], rv[1]) for rv in sorted(six.iteritems(self.handlers)) ]

    def get_enum(self):
        return self.enum

    def get_urls(self):
        return self.urls




from django.shortcuts import render
from rest_framework.response import Response
from rest_framework import serializers, status
from rest_framework.views import APIView
from djsw_wrapper.utils import LazyClass

SwaggerViewClass = APIView

# make http handler
# TODO: rewrite to class?
def SwaggerRequestMethodMaker(model = None):
    def model_handler(self, request, *args, **kwargs):
        data = kwargs.get('data', None)
        resp = model

        return Response(resp, status = status.HTTP_200_OK)

    def empty_handler(self, request, *args, **kwargs):
        data = kwargs.get('data', None)
        resp = None

        return Response(resp, status = status.HTTP_200_OK)

    if model:
        return model_handler
    else:
        return empty_handler

# make named APIView class with specified methods
class SwaggerViewMaker(LazyClass):
    oftype = SwaggerViewClass

    def as_view(self):
        c = self.as_class()

        return c.as_view()

class SwaggerRequestSerializerMaker(LazyClass):
    oftype = serializers.Serializer

# TODO: make proper model mapping
class SwaggerDataSerializerMaker(LazyClass):
    def __init__(self, *args, **kwargs):
        super(SwaggerDataSerializerMaker, self).__init__(*args, **kwargs)


from django.utils import six
from djsw_wrapper.router import SwaggerRouter
from djsw_wrapper.errors import SwaggerValidationError, SwaggerGenericError

import flex
import os
import re

class Swagger():
    # handle is local filename, file object, string or url
    def __init__(self, handle, module):
        self.schema = None
        self.module = None
        self.loaded = False
        self.handle = handle
        self.models = []
        self.router = None
        self.models = dict()

        # parse
        # TODO: proper errors
        try:
            self.schema = flex.load(self.handle)
            self.module = module
            self.loaded = True
        except:
            raise SwaggerGenericError('Cannot process schema {} : check resource availability'.format(self.handle))

        # make models for definitions
        if 'definitions' in self.schema:
            # make external models
            for name, data in six.iteritems(self.schema['definitions']):
                model = None

                if 'properties' in data:
                    model = list() #dict()

                    for prop, data in six.iteritems(data['properties']):
                        model.append(prop)

                if model:
                    self.models[name] = model

        # make routes
        if 'paths' in self.schema and 'basePath' in self.schema:
            self.router = SwaggerRouter(self.schema, self.module, self.models)
        else:
            raise SwaggerValidationError('Schema is missing paths and/or basePath values')

    # some advanced parsing techniques to be implemented
    def get_schema(self):
        if self.loaded:
            return self.schema
        else:
            raise SwaggerGenericError('You should load spec file first')

    def get_models(self):
        if self.loaded:
            return self.models
        else:
            raise SwaggerGenericError('You should load spec file first')



import os
import re
from jinja2 import Environment, FileSystemLoader

class _Singleton(type):
    """ A metaclass that creates a Singleton base class when called. """
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class Singleton(_Singleton('SingletonMeta', (object,), {})): pass

# creates class from dict on demand
class LazyClass(object):
    oftype = None

    def __init__(self, name = None, oftype = None):
        if not self.oftype and oftype:
            self.oftype = oftype

        assert name is not None, ('You should provide a name for new class')
        assert self.oftype is not None, ('You should provide a type for new class')

        self.handle = None
        self.attrs = dict()
        self.ready = False
        self.name = name

    def setup(self):
        self.handle = type(self.name, (self.oftype,), dict(self.attrs))
        self.ready = True

    def set_attr(self, key, value):
        self.attrs[key] = value

    def as_class(self):
        if not self.ready:
            self.setup()

        return self.handle

class Template():
    def __init__(self):
        self.loader = FileSystemLoader(os.path.join(os.path.dirname(__file__), 'templates'))
        self.env = Environment(loader = self.loader)

    def render(self, template_name, **kwargs):
        template = self.env.get_template(template_name)
        return template.render(**kwargs)

# processes $ref links
def Resolver(obj, path, full = False):
    m = re.search('#/(.*)/(.*)', path)
    x = None

    if full:
        b = obj[m.group(1)]
        x = b[m.group(2)]
    else:
        x = m.group(2)

    return x



from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from djsw_wrapper.core import Swagger

import logging

# Get an instance of a logger
logger = logging.getLogger(__name__)

class djswconfig(AppConfig):
    name = 'djsw_wrapper'
    verbose_name = "Django Swagger Wrapper"

    module = None
    schema = None
    swagger = None

    # startup
    def ready(self):
        self.schema = getattr(settings, 'SWAGGER_SCHEMA', None)

        if not self.schema:
            raise ImproperlyConfigured('You have to provide SWAGGER_SCHEMA setting pointing to desired schema')
        else:
            self.module = getattr(settings, 'SWAGGER_MODULE', None)
            self.swagger = Swagger(self.schema, self.module)



from django.utils.six import iteritems
from rest_framework import serializers

from djsw_wrapper.errors import SwaggerParameterError
from djsw_wrapper.makers import SwaggerRequestSerializerMaker


# TODO: rewrite to proper enum
class ParameterType():
    String = 0
    Number = 1
    Integer = 2
    Boolean = 3
    Array = 4
    Enum = 5
    File = 6

    typeset = {
        'string' : String,
        'number' : Number,
        'integer' : Integer,
        'boolean' : Boolean,
        'array' : Array,
        'enum' : Enum,
        'file' : File
    }

    oftype = None

    def get_type(self):
        return self.oftype

    def __init__(self, initial = None):
        if type(initial) is None:
            self.oftype = self.typeset.string
        elif type(initial) is str and initial.lower() in self.typeset.keys():
            self.oftype = self.typeset[initial.lower()]
        elif isinstance(initial, ParameterType):
            self.oftype = initial.get_type()
        else:
            raise SwaggerParameterError('Unknown parameter type: {0}'.format(string))

    def __repr__(self):
        return next(x for x, y in iteritems(self.typeset) if y == self.oftype)

class ParameterLocation():
    Query = 0
    Header = 1
    Path = 2
    FormData = 3
    Body = 4

    @staticmethod
    def fromString(string):
        if string.lower() == 'query':
            return ParameterLocation.Query
        elif string.lower() == 'header':
            return ParameterLocation.Header
        elif string.lower() == 'path':
            return ParameterLocation.Path
        elif string.lower() == 'formdata':
            return ParameterLocation.FormData
        elif string.lower() == 'body':
            return ParameterLocation.Body
        else:
            raise SwaggerParameterError('Unknown parameter location: {0}'.format(string))

class SwaggerParameter():

    def typemap(self, p):
        mapping = {
            ParameterType.String : serializers.CharField,
            ParameterType.Number : serializers.FloatField,
            ParameterType.Integer : serializers.IntegerField,
            ParameterType.Boolean : serializers.BooleanField,
            ParameterType.Array : serializers.ListField,
            ParameterType.Enum : serializers.ChoiceField,
            ParameterType.File : serializers.FileField
        }

        return mapping.get(p, None)

    # TODO: properly handle array and enums
    def __init__(self, schema):
        self._name = schema['name']
        self._enum = schema.get('enum', None)
        self._items = schema.get('items', None)
        self._oftype = ParameterType(schema['type']).get_type()
        self._location = ParameterLocation.fromString(schema['in'])
        self._required = schema.get('required', False)

        # default params
        self._params = { 'required' : self.required }

        # quick check for array
        if self._oftype == ParameterType.Array and 'items' not in schema:
            raise SwaggerParameterError('You should provide items dictionary for using array type')

        # quick check for file
        if self._oftype == ParameterType.File and self._location is not ParameterLocation.FormData:
            raise SwaggerParameterError('You have to use formData location for using file type')

        if self._oftype == ParameterType.Array:
            child = self.typemap(ParameterType(self._items['type']).get_type())
            self._params = { 'child': child() }

        if self._enum:
            self._oftype = ParameterType.Enum
            self._params = { 'choices' : self._enum }


    @property
    def name(self):
        return self._name

    @property
    def oftype(self):
        return self._oftype

    @property
    def location(self):
        return self._location

    @property
    def required(self):
        return self._required


    def __repr__(self):
        return "{} ({},{})".format(self._name, self._oftype, self._required)

    # TODO: properly handle array and enums
    # TODO: store serializer params in settings
    def as_field(self):
        items = None

        if self._oftype is ParameterType.String:
            self._params['max_length'] = 255 # to be discussed
        elif self._oftype is ParameterType.Enum:
            pass
        elif self._oftype is ParameterType.Number:
            #self._params['max_digits'] = 16
            #self._params['decimal_places'] = 4
            pass
        elif self._oftype is ParameterType.Array and items:
            pass

        field = self.typemap(self._oftype)

        # maybe exception?
        return field(**self._params) if field is not None else None

# automatically validates the data
def SwaggerRequestHandler(view, handler, params, *args, **kwargs):

    # wrapped request handler
    class SwaggerValidator(object):
        def __init__(self, view = None, serializer = None, func = None, params = None):
            self.serializer = serializer
            self.params = params
            self.func = func
            self.view = view

        # extract params with respect to their location
        def extract(self, request, uparams):
            data = dict()

            # TODO: what about JSON?
            for param in self.params:
                store = None
                value = None

                p = None
                n = param.name
                t = param.oftype

                if param.location == ParameterLocation.Query:
                    store = request.query_params
                elif param.location == ParameterLocation.Path:
                    store = uparams
                elif param.location in [ParameterLocation.FormData, ParameterLocation.Body]:
                    store = request.data

                if param.oftype == ParameterType.Array:
                    value = store.getlist(param.name, None)
                else:
                    value = store.get(param.name, None)

                if value:
                    data[param.name] = value

            return data

        # validate request data
        def process(self, request, *args, **kwargs):
            s_object = self.serializer.as_class()
            serializer = s_object(data = self.extract(request, kwargs))

            if serializer.is_valid(raise_exception = True):
                return self.func(self.view, request, data = serializer.data, *args, **kwargs)
            else:
                pass # s.errors contain detailed error

    # validate or not
    if not params:
        return handler
    else:
        serializer = SwaggerRequestSerializerMaker('SwaggerRequestSerializer')

        for param in params:
            serializer.set_attr(param.name, param.as_field())

        validator = SwaggerValidator(view, serializer, handler, params)

        return validator.process




from distutils.core import setup

with open('requirements.txt') as req:
    content = req.readlines()

setup(
  name = 'djsw_wrapper',
  packages = ['djsw_wrapper'],
  version = '0.1.9.8',
  description = 'Allows to build REST API directly from Swagger schema',
  author = 'Alex Revin',
  author_email = 'lyssdod@gmail.com',
  url = 'https://github.com/ErintLabs/django-openapi-gen',
  download_url = 'https://github.com/ErintLabs/django-openapi-gen/archive/0.1.9.8.tar.gz',
  install_requires = [x.strip() for x in content],
  keywords = ['django', 'swagger', 'schema', 'django-rest-framework'],
  classifiers = [],
)
